#include <WiFi.h>
#include <ESPmDNS.h>
#include <ArduinoOTA.h>

/* Nom du réseau WiFi */
const char ssid[] = "**********";
/* Mot de passe du réseau WiFi */
const char pass[] = "**********";

/* Mot de passe pour authentifier le téléversement via OTA */
const char OTAPass[] = "locoduino";

#define DEBUG

#ifdef DEBUG
#define DSB(v) Serial.begin(v)
#define DP(m) Serial.print(m)
#define DPLN(m) Serial.println(m)
#else
#define DSB(v)
#define DP(m)
#define DPLN(m)
#endif

class Connection {
public:
  typedef enum { OFFLINE, WIFI_STBY, WIFI_OK, MDNS_OK, OTA_OK } State;

private:
  static State sState;
  static uint32_t sLastDate;
  static uint32_t sPeriod;
  static char *sSsid;
  static char *sPass;
  static char *sName;
  static char *sOTAPass;

  Connection() {}
  
  static void update();
  static void initOTA();

#ifdef DEBUG
  static void startingOTA();
  static void OTAInProgress(unsigned int inProgress, unsigned int inTotal);
  static void endingOTA();
  static void OTAError(ota_error_t error);
#endif
  

public:
  static void setup(const char *inSsid,
                    const char *inPass,
                    const char *inName,
                    const char *inOTAPass = NULL,
                    const uint32_t inPeriod = 1000ul);
  static void loop();
  static bool isOnline();
};

Connection::State Connection::sState = OFFLINE;
uint32_t Connection::sLastDate = 0ul;
uint32_t Connection::sPeriod = 1000ul;
char *Connection::sSsid = NULL;
char *Connection::sPass = NULL;
char *Connection::sName = NULL;
char *Connection::sOTAPass = NULL;

void Connection::update()
{
  switch (sState) {
    case OFFLINE:
      if (sSsid != NULL && sPass != NULL) {
        DP("Connexion à : ");
        WiFi.begin(sSsid, sPass);
        sState = WIFI_STBY;
      }
      break;
      
    case WIFI_STBY:
      if (WiFi.status() == WL_CONNECTED) {
        DP(' ');
        DPLN(sSsid);
        sState = WIFI_OK;
      } else {
        DP('.');
      }
      break;
      
    case WIFI_OK:
      if (WiFi.status() == WL_CONNECTED) {
        if (sName != NULL) {
          MDNS.begin(sName);
          DP("Demarrage de mDNS sous le nom ");
          DP(sName);
          DPLN(".local");
          sState = MDNS_OK;
        }
      } else {
        DPLN("Perte de connexion");
        sState = OFFLINE;
      }
      break;

    case MDNS_OK:
      if (WiFi.status() == WL_CONNECTED) {
        initOTA();
        sState = OTA_OK;
      } else {
        DPLN("Perte de connexion");
        sState = OFFLINE;
      }
      break;

    case OTA_OK:
      if (WiFi.status() != WL_CONNECTED) {
        DPLN("Perte de connexion");
        ArduinoOTA.end();
        sState = OFFLINE;
      }
      break;
  }
}

void Connection::initOTA()
{
  DP("Demarrage OTA sous le nom ");
  DPLN(sName);
  ArduinoOTA.setHostname(sName);
  if (sOTAPass != NULL) {
    ArduinoOTA.setPassword(sOTAPass);
  }
  
#ifdef DEBUG
  ArduinoOTA.onStart(startingOTA);
  ArduinoOTA.onProgress(OTAInProgress);
  ArduinoOTA.onEnd(endingOTA);
  ArduinoOTA.onError(OTAError);
#endif

  ArduinoOTA.begin();
}

#ifdef DEBUG
void Connection::startingOTA()
{
  const int command = ArduinoOTA.getCommand();
  if (command == U_FLASH) {
    DPLN("Téléversement du firmware");
  } else {
    DP("Commande inattendue ");
    DPLN(command);
  }
}

void Connection::OTAInProgress(unsigned int inProgress, unsigned int inTotal)
{
  static uint32_t previousNumberOfDots = 0;
  const uint32_t numberOfDots = 50 * inProgress / inTotal;
  for (uint32_t dot = previousNumberOfDots; dot < numberOfDots; dot++) {
    DP('.');
  }
  if (inTotal == inProgress) {
    DPLN();
  }
  previousNumberOfDots = numberOfDots;
}

void Connection::endingOTA()
{
  DPLN("Téléversement terminé");
}

void Connection::OTAError(ota_error_t error)
{
  DP("Erreur[");
  DP(error);
  DP("] : ");
  switch (error) {
    case OTA_AUTH_ERROR:    DPLN("L'authentification a échoué"); break;
    case OTA_BEGIN_ERROR:   DPLN("Échec au début");              break;
    case OTA_CONNECT_ERROR: DPLN("Échec à la connexion");        break;
    case OTA_RECEIVE_ERROR: DPLN("Échec à la réception");        break;
    case OTA_END_ERROR:     DPLN("Échec à la fermeture");        break;
  }
}
#endif


void Connection::setup(const char *inSsid,
                       const char *inPass,
                       const char *inName,
                       const char *inOTAPass,
                       const uint32_t inPeriod)
{
  sPeriod = inPeriod;
  sSsid = (char *)inSsid;
  sPass = (char *)inPass;
  sName = (char *)inName;
  sOTAPass = (char *)inOTAPass;
}

void Connection::loop()
{
  const uint32_t currentDate = millis();
  if ((currentDate - sLastDate) >= sPeriod) {
    update();
    sLastDate = currentDate;
  }
  ArduinoOTA.handle();
}

bool Connection::isOnline()
{ 
  return (sState == OTA_OK);
}

void setup() {
  /* Pour le debug */
  DSB(115200);
  /* Connexion */
  Connection::setup(ssid, pass, "Eclairage", OTAPass);
}

void loop() {
  Connection::loop();
}
